@@ -3,39 +3,19 @@ from __future__ import print_function
 
 import argparse
 import collections
-from functools import wraps
+from functools import wraps, partial
+import inspect
 import re
 import sys
+import warnings
 
 
-__version__ = "2.9.1"
+__version__ = "2.10.0"
 __author__ = "Kostiantyn Rybnikov"
 __author_email__ = "k-bx@k-bx.com"
 __maintainer__ = ["Sebastien Celles", "Tom Schraitle"]
 __maintainer_email__ = "s.celles@gmail.com"
 
-_REGEX = re.compile(
-    r"""
-        ^
-        (?P<major>0|[1-9]\d*)
-        \.
-        (?P<minor>0|[1-9]\d*)
-        \.
-        (?P<patch>0|[1-9]\d*)
-        (?:-(?P<prerelease>
-            (?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)
-            (?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*
-        ))?
-        (?:\+(?P<build>
-            [0-9a-zA-Z-]+
-            (?:\.[0-9a-zA-Z-]+)*
-        ))?
-        $
-        """,
-    re.VERBOSE,
-)
-
-_LAST_NUMBER = re.compile(r"(?:[^\d]*(\d+)[^\d]*)+")
 
 #: Contains the implemented semver.org version of the spec
 SEMVER_SPEC_VERSION = "2.0.0"
@@ -44,13 +24,68 @@ SEMVER_SPEC_VERSION = "2.0.0"
 if not hasattr(__builtins__, "cmp"):
 
     def cmp(a, b):
+        """Return negative if a<b, zero if a==b, positive if a>b."""
         return (a > b) - (a < b)
 
 
+def deprecated(func=None, replace=None, version=None, category=DeprecationWarning):
+    """
+    Decorates a function to output a deprecation warning.
+
+    :param func: the function to decorate (or None)
+    :param str replace: the function to replace (use the full qualified
+        name like ``semver.VersionInfo.bump_major``.
+    :param str version: the first version when this function was deprecated.
+    :param category: allow you to specify the deprecation warning class
+        of your choice. By default, it's  :class:`DeprecationWarning`, but
+        you can choose :class:`PendingDeprecationWarning` or a custom class.
+    """
+
+    if func is None:
+        return partial(deprecated, replace=replace, version=version, category=category)
+
+    @wraps(func)
+    def wrapper(*args, **kwargs):
+        msg = ["Function '{m}.{f}' is deprecated."]
+
+        if version:
+            msg.append("Deprecated since version {v}. ")
+        msg.append("This function will be removed in semver 3.")
+        if replace:
+            msg.append("Use {r!r} instead.")
+        else:
+            msg.append("Use the respective 'semver.VersionInfo.{r}' instead.")
+
+        # hasattr is needed for Python2 compatibility:
+        f = func.__qualname__ if hasattr(func, "__qualname__") else func.__name__
+        r = replace or f
+
+        frame = inspect.currentframe().f_back
+
+        msg = " ".join(msg)
+        warnings.warn_explicit(
+            msg.format(m=func.__module__, f=f, r=r, v=version),
+            category=category,
+            filename=inspect.getfile(frame.f_code),
+            lineno=frame.f_lineno,
+        )
+        # As recommended in the Python documentation
+        # https://docs.python.org/3/library/inspect.html#the-interpreter-stack
+        # better remove the interpreter stack:
+        del frame
+        return func(*args, **kwargs)
+
+    return wrapper
+
+
+@deprecated(version="2.10.0")
 def parse(version):
     """
     Parse version to major, minor, patch, pre-release, build parts.
 
+    .. deprecated:: 2.10.0
+       Use :func:`semver.VersionInfo.parse` instead.
+
     :param version: version string
     :return: dictionary with the keys 'build', 'major', 'minor', 'patch',
              and 'prerelease'. The prerelease or build keys can be None
@@ -69,17 +104,7 @@ def parse(version):
     >>> ver['build']
     'build.4'
     """
-    match = _REGEX.match(version)
-    if match is None:
-        raise ValueError("%s is not valid SemVer string" % version)
-
-    version_parts = match.groupdict()
-
-    version_parts["major"] = int(version_parts["major"])
-    version_parts["minor"] = int(version_parts["minor"])
-    version_parts["patch"] = int(version_parts["patch"])
-
-    return version_parts
+    return VersionInfo.parse(version).to_dict()
 
 
 def comparator(operator):
@@ -87,7 +112,7 @@ def comparator(operator):
 
     @wraps(operator)
     def wrapper(self, other):
-        comparable_types = (VersionInfo, dict, tuple)
+        comparable_types = (VersionInfo, dict, tuple, list, str)
         if not isinstance(other, comparable_types):
             raise TypeError(
                 "other type %r must be in %r" % (type(other), comparable_types)
@@ -110,6 +135,29 @@ class VersionInfo(object):
     """
 
     __slots__ = ("_major", "_minor", "_patch", "_prerelease", "_build")
+    #: Regex for number in a prerelease
+    _LAST_NUMBER = re.compile(r"(?:[^\d]*(\d+)[^\d]*)+")
+    #: Regex for a semver version
+    _REGEX = re.compile(
+        r"""
+            ^
+            (?P<major>0|[1-9]\d*)
+            \.
+            (?P<minor>0|[1-9]\d*)
+            \.
+            (?P<patch>0|[1-9]\d*)
+            (?:-(?P<prerelease>
+                (?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)
+                (?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*
+            ))?
+            (?:\+(?P<build>
+                [0-9a-zA-Z-]+
+                (?:\.[0-9a-zA-Z-]+)*
+            ))?
+            $
+        """,
+        re.VERBOSE,
+    )
 
     def __init__(self, major, minor=0, patch=0, prerelease=None, build=None):
         self._major = int(major)
@@ -120,7 +168,7 @@ class VersionInfo(object):
 
     @property
     def major(self):
-        """The major part of a version."""
+        """The major part of a version (read-only)."""
         return self._major
 
     @major.setter
@@ -129,7 +177,7 @@ class VersionInfo(object):
 
     @property
     def minor(self):
-        """The minor part of a version."""
+        """The minor part of a version (read-only)."""
         return self._minor
 
     @minor.setter
@@ -138,7 +186,7 @@ class VersionInfo(object):
 
     @property
     def patch(self):
-        """The patch part of a version."""
+        """The patch part of a version (read-only)."""
         return self._patch
 
     @patch.setter
@@ -147,7 +195,7 @@ class VersionInfo(object):
 
     @property
     def prerelease(self):
-        """The prerelease part of a version."""
+        """The prerelease part of a version (read-only)."""
         return self._prerelease
 
     @prerelease.setter
@@ -156,17 +204,45 @@ class VersionInfo(object):
 
     @property
     def build(self):
-        """The build part of a version."""
+        """The build part of a version (read-only)."""
         return self._build
 
     @build.setter
     def build(self, value):
         raise AttributeError("attribute 'build' is readonly")
 
-    def _astuple(self):
+    def to_tuple(self):
+        """
+        Convert the VersionInfo object to a tuple.
+
+        .. versionadded:: 2.10.0
+           Renamed ``VersionInfo._astuple`` to ``VersionInfo.to_tuple`` to
+           make this function available in the public API.
+
+        :return: a tuple with all the parts
+        :rtype: tuple
+
+        >>> semver.VersionInfo(5, 3, 1).to_tuple()
+        (5, 3, 1, None, None)
+        """
         return (self.major, self.minor, self.patch, self.prerelease, self.build)
 
-    def _asdict(self):
+    def to_dict(self):
+        """
+        Convert the VersionInfo object to an OrderedDict.
+
+        .. versionadded:: 2.10.0
+           Renamed ``VersionInfo._asdict`` to ``VersionInfo.to_dict`` to
+           make this function available in the public API.
+
+        :return: an OrderedDict with the keys in the order ``major``, ``minor``,
+          ``patch``, ``prerelease``, and ``build``.
+        :rtype: :class:`collections.OrderedDict`
+
+        >>> semver.VersionInfo(3, 2, 1).to_dict()
+        OrderedDict([('major', 3), ('minor', 2), ('patch', 1), \
+('prerelease', None), ('build', None)])
+        """
         return collections.OrderedDict(
             (
                 ("major", self.major),
@@ -177,12 +253,43 @@ class VersionInfo(object):
             )
         )
 
+    # For compatibility reasons:
+    @deprecated(replace="semver.VersionInfo.to_tuple", version="2.10.0")
+    def _astuple(self):
+        return self.to_tuple()  # pragma: no cover
+
+    _astuple.__doc__ = to_tuple.__doc__
+
+    @deprecated(replace="semver.VersionInfo.to_dict", version="2.10.0")
+    def _asdict(self):
+        return self.to_dict()  # pragma: no cover
+
+    _asdict.__doc__ = to_dict.__doc__
+
     def __iter__(self):
         """Implement iter(self)."""
         # As long as we support Py2.7, we can't use the "yield from" syntax
-        for v in self._astuple():
+        for v in self.to_tuple():
             yield v
 
+    @staticmethod
+    def _increment_string(string):
+        """
+        Look for the last sequence of number(s) in a string and increment.
+
+        :param str string: the string to search for.
+        :return: the incremented string
+
+        Source:
+        http://code.activestate.com/recipes/442460-increment-numbers-in-a-string/#c1
+        """
+        match = VersionInfo._LAST_NUMBER.search(string)
+        if match:
+            next_ = str(int(match.group(1)) + 1)
+            start, end = match.span(1)
+            string = string[: max(end - len(next_), start)] + next_ + string[end:]
+        return string
+
     def bump_major(self):
         """
         Raise the major part of the version, return a new object but leave self
@@ -191,11 +298,12 @@ class VersionInfo(object):
         :return: new object with the raised major part
         :rtype: VersionInfo
 
-        >>> ver = semver.parse_version_info("3.4.5")
+        >>> ver = semver.VersionInfo.parse("3.4.5")
         >>> ver.bump_major()
         VersionInfo(major=4, minor=0, patch=0, prerelease=None, build=None)
         """
-        return parse_version_info(bump_major(str(self)))
+        cls = type(self)
+        return cls(self._major + 1)
 
     def bump_minor(self):
         """
@@ -205,11 +313,12 @@ class VersionInfo(object):
         :return: new object with the raised minor part
         :rtype: VersionInfo
 
-        >>> ver = semver.parse_version_info("3.4.5")
+        >>> ver = semver.VersionInfo.parse("3.4.5")
         >>> ver.bump_minor()
         VersionInfo(major=3, minor=5, patch=0, prerelease=None, build=None)
         """
-        return parse_version_info(bump_minor(str(self)))
+        cls = type(self)
+        return cls(self._major, self._minor + 1)
 
     def bump_patch(self):
         """
@@ -219,11 +328,12 @@ class VersionInfo(object):
         :return: new object with the raised patch part
         :rtype: VersionInfo
 
-        >>> ver = semver.parse_version_info("3.4.5")
+        >>> ver = semver.VersionInfo.parse("3.4.5")
         >>> ver.bump_patch()
         VersionInfo(major=3, minor=4, patch=6, prerelease=None, build=None)
         """
-        return parse_version_info(bump_patch(str(self)))
+        cls = type(self)
+        return cls(self._major, self._minor, self._patch + 1)
 
     def bump_prerelease(self, token="rc"):
         """
@@ -234,12 +344,14 @@ class VersionInfo(object):
         :return: new object with the raised prerelease part
         :rtype: str
 
-        >>> ver = semver.parse_version_info("3.4.5-rc.1")
+        >>> ver = semver.VersionInfo.parse("3.4.5-rc.1")
         >>> ver.bump_prerelease()
         VersionInfo(major=3, minor=4, patch=5, prerelease='rc.2', \
 build=None)
         """
-        return parse_version_info(bump_prerelease(str(self), token))
+        cls = type(self)
+        prerelease = cls._increment_string(self._prerelease or (token or "rc") + ".0")
+        return cls(self._major, self._minor, self._patch, prerelease)
 
     def bump_build(self, token="build"):
         """
@@ -250,46 +362,248 @@ build=None)
         :return: new object with the raised build part
         :rtype: str
 
-        >>> ver = semver.parse_version_info("3.4.5-rc.1+build.9")
+        >>> ver = semver.VersionInfo.parse("3.4.5-rc.1+build.9")
         >>> ver.bump_build()
         VersionInfo(major=3, minor=4, patch=5, prerelease='rc.1', \
 build='build.10')
         """
-        return parse_version_info(bump_build(str(self), token))
+        cls = type(self)
+        build = cls._increment_string(self._build or (token or "build") + ".0")
+        return cls(self._major, self._minor, self._patch, self._prerelease, build)
+
+    def compare(self, other):
+        """
+        Compare self with other.
+
+        :param other: the second version (can be string, a dict, tuple/list, or
+             a VersionInfo instance)
+        :return: The return value is negative if ver1 < ver2,
+             zero if ver1 == ver2 and strictly positive if ver1 > ver2
+        :rtype: int
+
+        >>> semver.VersionInfo.parse("1.0.0").compare("2.0.0")
+        -1
+        >>> semver.VersionInfo.parse("2.0.0").compare("1.0.0")
+        1
+        >>> semver.VersionInfo.parse("2.0.0").compare("2.0.0")
+        0
+        >>> semver.VersionInfo.parse("2.0.0").compare(dict(major=2, minor=0, patch=0))
+        0
+        """
+        cls = type(self)
+        if isinstance(other, str):
+            other = cls.parse(other)
+        elif isinstance(other, dict):
+            other = cls(**other)
+        elif isinstance(other, (tuple, list)):
+            other = cls(*other)
+        elif not isinstance(other, cls):
+            raise TypeError(
+                "Expected str or {} instance, but got {}".format(
+                    cls.__name__, type(other)
+                )
+            )
+
+        v1 = self.to_tuple()[:3]
+        v2 = other.to_tuple()[:3]
+        x = cmp(v1, v2)
+        if x:
+            return x
+
+        rc1, rc2 = self.prerelease, other.prerelease
+        rccmp = _nat_cmp(rc1, rc2)
+
+        if not rccmp:
+            return 0
+        if not rc1:
+            return 1
+        elif not rc2:
+            return -1
+
+        return rccmp
+
+    def next_version(self, part, prerelease_token="rc"):
+        """
+        Determines next version, preserving natural order.
+
+        .. versionadded:: 2.10.0
+
+        This function is taking prereleases into account.
+        The "major", "minor", and "patch" raises the respective parts like
+        the ``bump_*`` functions. The real difference is using the
+        "preprelease" part. It gives you the next patch version of the prerelease,
+        for example:
+
+        >>> str(semver.VersionInfo.parse("0.1.4").next_version("prerelease"))
+        '0.1.5-rc.1'
+
+        :param part: One of "major", "minor", "patch", or "prerelease"
+        :param prerelease_token: prefix string of prerelease, defaults to 'rc'
+        :return:
+        """
+        validparts = {
+            "major",
+            "minor",
+            "patch",
+            "prerelease",
+            # "build", # currently not used
+        }
+        if part not in validparts:
+            raise ValueError(
+                "Invalid part. Expected one of {validparts}, but got {part!r}".format(
+                    validparts=validparts, part=part
+                )
+            )
+        version = self
+        if (version.prerelease or version.build) and (
+            part == "patch"
+            or (part == "minor" and version.patch == 0)
+            or (part == "major" and version.minor == version.patch == 0)
+        ):
+            return version.replace(prerelease=None, build=None)
+
+        if part in ("major", "minor", "patch"):
+            return str(getattr(version, "bump_" + part)())
+
+        if not version.prerelease:
+            version = version.bump_patch()
+        return version.bump_prerelease(prerelease_token)
 
     @comparator
     def __eq__(self, other):
-        return _compare_by_keys(self._asdict(), _to_dict(other)) == 0
+        return self.compare(other) == 0
 
     @comparator
     def __ne__(self, other):
-        return _compare_by_keys(self._asdict(), _to_dict(other)) != 0
+        return self.compare(other) != 0
 
     @comparator
     def __lt__(self, other):
-        return _compare_by_keys(self._asdict(), _to_dict(other)) < 0
+        return self.compare(other) < 0
 
     @comparator
     def __le__(self, other):
-        return _compare_by_keys(self._asdict(), _to_dict(other)) <= 0
+        return self.compare(other) <= 0
 
     @comparator
     def __gt__(self, other):
-        return _compare_by_keys(self._asdict(), _to_dict(other)) > 0
+        return self.compare(other) > 0
 
     @comparator
     def __ge__(self, other):
-        return _compare_by_keys(self._asdict(), _to_dict(other)) >= 0
+        return self.compare(other) >= 0
+
+    def __getitem__(self, index):
+        """
+        self.__getitem__(index) <==> self[index]
+
+        Implement getitem. If the part requested is undefined, or a part of the
+        range requested is undefined, it will throw an index error.
+        Negative indices are not supported
+
+        :param Union[int, slice] index: a positive integer indicating the
+               offset or a :func:`slice` object
+        :raises: IndexError, if index is beyond the range or a part is None
+        :return: the requested part of the version at position index
+
+        >>> ver = semver.VersionInfo.parse("3.4.5")
+        >>> ver[0], ver[1], ver[2]
+        (3, 4, 5)
+        """
+        if isinstance(index, int):
+            index = slice(index, index + 1)
+
+        if (
+            isinstance(index, slice)
+            and (index.start is None or index.start < 0)
+            and (index.stop is None or index.stop < 0)
+        ):
+            raise IndexError("Version index cannot be negative")
+
+        # Could raise IndexError:
+        part = tuple(filter(None, self.to_tuple()[index]))
+
+        if len(part) == 1:
+            part = part[0]
+        if not part:
+            raise IndexError("Version part undefined")
+        return part
 
     def __repr__(self):
-        s = ", ".join("%s=%r" % (key, val) for key, val in self._asdict().items())
+        s = ", ".join("%s=%r" % (key, val) for key, val in self.to_dict().items())
         return "%s(%s)" % (type(self).__name__, s)
 
     def __str__(self):
-        return format_version(*(self._astuple()))
+        """str(self)"""
+        version = "%d.%d.%d" % (self.major, self.minor, self.patch)
+        if self.prerelease:
+            version += "-%s" % self.prerelease
+        if self.build:
+            version += "+%s" % self.build
+        return version
 
     def __hash__(self):
-        return hash(self._astuple())
+        return hash(self.to_tuple())
+
+    def finalize_version(self):
+        """
+        Remove any prerelease and build metadata from the version.
+
+        :return: a new instance with the finalized version string
+        :rtype: :class:`VersionInfo`
+
+        >>> str(semver.VersionInfo.parse('1.2.3-rc.5').finalize_version())
+        '1.2.3'
+        """
+        cls = type(self)
+        return cls(self.major, self.minor, self.patch)
+
+    def match(self, match_expr):
+        """
+        Compare self to match a match expression.
+
+        :param str match_expr: operator and version; valid operators are
+              <   smaller than
+              >   greater than
+              >=  greator or equal than
+              <=  smaller or equal than
+              ==  equal
+              !=  not equal
+        :return: True if the expression matches the version, otherwise False
+        :rtype: bool
+
+        >>> semver.VersionInfo.parse("2.0.0").match(">=1.0.0")
+        True
+        >>> semver.VersionInfo.parse("1.0.0").match(">1.0.0")
+        False
+        """
+        prefix = match_expr[:2]
+        if prefix in (">=", "<=", "==", "!="):
+            match_version = match_expr[2:]
+        elif prefix and prefix[0] in (">", "<"):
+            prefix = prefix[0]
+            match_version = match_expr[1:]
+        else:
+            raise ValueError(
+                "match_expr parameter should be in format <op><ver>, "
+                "where <op> is one of "
+                "['<', '>', '==', '<=', '>=', '!=']. "
+                "You provided: %r" % match_expr
+            )
+
+        possibilities_dict = {
+            ">": (1,),
+            "<": (-1,),
+            "==": (0,),
+            "!=": (-1, 1),
+            ">=": (0, 1),
+            "<=": (-1, 0),
+        }
+
+        possibilities = possibilities_dict[prefix]
+        cmp_res = self.compare(match_version)
+
+        return cmp_res in possibilities
 
     @staticmethod
     def parse(version):
@@ -298,30 +612,45 @@ build='build.10')
 
         :param version: version string
         :return: a :class:`semver.VersionInfo` instance
+        :raises: :class:`ValueError`
         :rtype: :class:`semver.VersionInfo`
 
         >>> semver.VersionInfo.parse('3.4.5-pre.2+build.4')
         VersionInfo(major=3, minor=4, patch=5, \
 prerelease='pre.2', build='build.4')
         """
-        return parse_version_info(version)
+        match = VersionInfo._REGEX.match(version)
+        if match is None:
+            raise ValueError("%s is not valid SemVer string" % version)
+
+        version_parts = match.groupdict()
+
+        version_parts["major"] = int(version_parts["major"])
+        version_parts["minor"] = int(version_parts["minor"])
+        version_parts["patch"] = int(version_parts["patch"])
+
+        return VersionInfo(**version_parts)
 
     def replace(self, **parts):
-        """Replace one or more parts of a version and return a new
-          :class:`semver.VersionInfo` object, but leave self untouched
+        """
+        Replace one or more parts of a version and return a new
+        :class:`semver.VersionInfo` object, but leave self untouched
+
+        .. versionadded:: 2.9.0
+           Added :func:`VersionInfo.replace`
 
         :param dict parts: the parts to be updated. Valid keys are:
           ``major``, ``minor``, ``patch``, ``prerelease``, or ``build``
         :return: the new :class:`semver.VersionInfo` object with the changed
           parts
-        :raises: TypeError, if ``parts`` contains invalid keys
+        :raises: :class:`TypeError`, if ``parts`` contains invalid keys
         """
-        version = self._asdict()
+        version = self.to_dict()
         version.update(parts)
         try:
             return VersionInfo(**version)
         except TypeError:
-            unknownkeys = set(parts) - set(self._asdict())
+            unknownkeys = set(parts) - set(self.to_dict())
             error = "replace() got %d unexpected keyword " "argument(s): %s" % (
                 len(unknownkeys),
                 ", ".join(unknownkeys),
@@ -333,6 +662,8 @@ prerelease='pre.2', build='build.4')
         """
         Check if the string is a valid semver version.
 
+        .. versionadded:: 2.9.1
+
         :param str version: the version string to check
         :return: True if the version string is a valid semver version, False
                  otherwise.
@@ -345,23 +676,22 @@ prerelease='pre.2', build='build.4')
             return False
 
 
-def _to_dict(obj):
-    if isinstance(obj, VersionInfo):
-        return obj._asdict()
-    elif isinstance(obj, tuple):
-        return VersionInfo(*obj)._asdict()
-    return obj
-
-
+@deprecated(replace="semver.VersionInfo.parse", version="2.10.0")
 def parse_version_info(version):
     """
     Parse version string to a VersionInfo instance.
 
+    .. deprecated:: 2.10.0
+       Use :func:`semver.VersionInfo.parse` instead.
+
+    .. versionadded:: 2.7.2
+       Added :func:`semver.parse_version_info`
+
     :param version: version string
     :return: a :class:`VersionInfo` instance
     :rtype: :class:`VersionInfo`
 
-    >>> version_info = semver.parse_version_info("3.4.5-pre.2+build.4")
+    >>> version_info = semver.VersionInfo.parse("3.4.5-pre.2+build.4")
     >>> version_info.major
     3
     >>> version_info.minor
@@ -373,16 +703,7 @@ def parse_version_info(version):
     >>> version_info.build
     'build.4'
     """
-    parts = parse(version)
-    version_info = VersionInfo(
-        parts["major"],
-        parts["minor"],
-        parts["patch"],
-        parts["prerelease"],
-        parts["build"],
-    )
-
-    return version_info
+    return VersionInfo.parse(version)
 
 
 def _nat_cmp(a, b):
@@ -412,28 +733,10 @@ def _nat_cmp(a, b):
         return cmp(len(a), len(b))
 
 
-def _compare_by_keys(d1, d2):
-    for key in ["major", "minor", "patch"]:
-        v = cmp(d1.get(key), d2.get(key))
-        if v:
-            return v
-
-    rc1, rc2 = d1.get("prerelease"), d2.get("prerelease")
-    rccmp = _nat_cmp(rc1, rc2)
-
-    if not rccmp:
-        return 0
-    if not rc1:
-        return 1
-    elif not rc2:
-        return -1
-
-    return rccmp
-
-
+@deprecated(version="2.10.0")
 def compare(ver1, ver2):
     """
-    Compare two versions.
+    Compare two versions strings.
 
     :param ver1: version string 1
     :param ver2: version string 2
@@ -448,15 +751,14 @@ def compare(ver1, ver2):
     >>> semver.compare("2.0.0", "2.0.0")
     0
     """
-
-    v1, v2 = parse(ver1), parse(ver2)
-
-    return _compare_by_keys(v1, v2)
+    v1 = VersionInfo.parse(ver1)
+    return v1.compare(ver2)
 
 
+@deprecated(version="2.10.0")
 def match(version, match_expr):
     """
-    Compare two versions through a comparison.
+    Compare two versions strings through a comparison.
 
     :param str version: a version string
     :param str match_expr: operator and version; valid operators are
@@ -474,38 +776,13 @@ def match(version, match_expr):
     >>> semver.match("1.0.0", ">1.0.0")
     False
     """
-    prefix = match_expr[:2]
-    if prefix in (">=", "<=", "==", "!="):
-        match_version = match_expr[2:]
-    elif prefix and prefix[0] in (">", "<"):
-        prefix = prefix[0]
-        match_version = match_expr[1:]
-    else:
-        raise ValueError(
-            "match_expr parameter should be in format <op><ver>, "
-            "where <op> is one of "
-            "['<', '>', '==', '<=', '>=', '!=']. "
-            "You provided: %r" % match_expr
-        )
-
-    possibilities_dict = {
-        ">": (1,),
-        "<": (-1,),
-        "==": (0,),
-        "!=": (-1, 1),
-        ">=": (0, 1),
-        "<=": (-1, 0),
-    }
-
-    possibilities = possibilities_dict[prefix]
-    cmp_res = compare(version, match_version)
-
-    return cmp_res in possibilities
+    ver = VersionInfo.parse(version)
+    return ver.match(match_expr)
 
 
 def max_ver(ver1, ver2):
     """
-    Returns the greater version of two versions.
+    Returns the greater version of two versions strings.
 
     :param ver1: version string 1
     :param ver2: version string 2
@@ -515,16 +792,20 @@ def max_ver(ver1, ver2):
     >>> semver.max_ver("1.0.0", "2.0.0")
     '2.0.0'
     """
-    cmp_res = compare(ver1, ver2)
-    if cmp_res == 0 or cmp_res == 1:
-        return ver1
+    if isinstance(ver1, str):
+        ver1 = VersionInfo.parse(ver1)
+    elif not isinstance(ver1, VersionInfo):
+        raise TypeError()
+    cmp_res = ver1.compare(ver2)
+    if cmp_res >= 0:
+        return str(ver1)
     else:
         return ver2
 
 
 def min_ver(ver1, ver2):
     """
-    Returns the smaller version of two versions.
+    Returns the smaller version of two versions strings.
 
     :param ver1: version string 1
     :param ver2: version string 2
@@ -534,16 +815,21 @@ def min_ver(ver1, ver2):
     >>> semver.min_ver("1.0.0", "2.0.0")
     '1.0.0'
     """
-    cmp_res = compare(ver1, ver2)
-    if cmp_res == 0 or cmp_res == -1:
-        return ver1
+    ver1 = VersionInfo.parse(ver1)
+    cmp_res = ver1.compare(ver2)
+    if cmp_res <= 0:
+        return str(ver1)
     else:
         return ver2
 
 
+@deprecated(replace="str(versionobject)", version="2.10.0")
 def format_version(major, minor, patch, prerelease=None, build=None):
     """
-    Format a version according to the Semantic Versioning specification.
+    Format a version string according to the Semantic Versioning specification.
+
+    .. deprecated:: 2.10.0
+       Use ``str(VersionInfo(VERSION)`` instead.
 
     :param int major: the required major part of a version
     :param int minor: the required minor part of a version
@@ -556,33 +842,16 @@ def format_version(major, minor, patch, prerelease=None, build=None):
     >>> semver.format_version(3, 4, 5, 'pre.2', 'build.4')
     '3.4.5-pre.2+build.4'
     """
-    version = "%d.%d.%d" % (major, minor, patch)
-    if prerelease is not None:
-        version = version + "-%s" % prerelease
-
-    if build is not None:
-        version = version + "+%s" % build
-
-    return version
-
-
-def _increment_string(string):
-    """
-    Look for the last sequence of number(s) in a string and increment, from:
-
-    http://code.activestate.com/recipes/442460-increment-numbers-in-a-string/#c1
-    """
-    match = _LAST_NUMBER.search(string)
-    if match:
-        next_ = str(int(match.group(1)) + 1)
-        start, end = match.span(1)
-        string = string[: max(end - len(next_), start)] + next_ + string[end:]
-    return string
+    return str(VersionInfo(major, minor, patch, prerelease, build))
 
 
+@deprecated(version="2.10.0")
 def bump_major(version):
     """
-    Raise the major part of the version.
+    Raise the major part of the version string.
+
+    .. deprecated:: 2.10.0
+       Use :func:`semver.VersionInfo.bump_major` instead.
 
     :param: version string
     :return: the raised version string
@@ -591,13 +860,16 @@ def bump_major(version):
     >>> semver.bump_major("3.4.5")
     '4.0.0'
     """
-    verinfo = parse(version)
-    return format_version(verinfo["major"] + 1, 0, 0)
+    return str(VersionInfo.parse(version).bump_major())
 
 
+@deprecated(version="2.10.0")
 def bump_minor(version):
     """
-    Raise the minor part of the version.
+    Raise the minor part of the version string.
+
+    .. deprecated:: 2.10.0
+       Use :func:`semver.VersionInfo.bump_minor` instead.
 
     :param: version string
     :return: the raised version string
@@ -606,13 +878,16 @@ def bump_minor(version):
     >>> semver.bump_minor("3.4.5")
     '3.5.0'
     """
-    verinfo = parse(version)
-    return format_version(verinfo["major"], verinfo["minor"] + 1, 0)
+    return str(VersionInfo.parse(version).bump_minor())
 
 
+@deprecated(version="2.10.0")
 def bump_patch(version):
     """
-    Raise the patch part of the version.
+    Raise the patch part of the version string.
+
+    .. deprecated:: 2.10.0
+       Use :func:`semver.VersionInfo.bump_patch` instead.
 
     :param: version string
     :return: the raised version string
@@ -621,13 +896,16 @@ def bump_patch(version):
     >>> semver.bump_patch("3.4.5")
     '3.4.6'
     """
-    verinfo = parse(version)
-    return format_version(verinfo["major"], verinfo["minor"], verinfo["patch"] + 1)
+    return str(VersionInfo.parse(version).bump_patch())
 
 
+@deprecated(version="2.10.0")
 def bump_prerelease(version, token="rc"):
     """
-    Raise the prerelease part of the version.
+    Raise the prerelease part of the version string.
+
+    .. deprecated:: 2.10.0
+       Use :func:`semver.VersionInfo.bump_prerelease` instead.
 
     :param version: version string
     :param token: defaults to 'rc'
@@ -637,18 +915,16 @@ def bump_prerelease(version, token="rc"):
     >>> semver.bump_prerelease('3.4.5', 'dev')
     '3.4.5-dev.1'
     """
-    verinfo = parse(version)
-    verinfo["prerelease"] = _increment_string(
-        verinfo["prerelease"] or (token or "rc") + ".0"
-    )
-    return format_version(
-        verinfo["major"], verinfo["minor"], verinfo["patch"], verinfo["prerelease"]
-    )
+    return str(VersionInfo.parse(version).bump_prerelease(token))
 
 
+@deprecated(version="2.10.0")
 def bump_build(version, token="build"):
     """
-    Raise the build part of the version.
+    Raise the build part of the version string.
+
+    .. deprecated:: 2.10.0
+       Use :func:`semver.VersionInfo.bump_build` instead.
 
     :param version: version string
     :param token: defaults to 'build'
@@ -658,20 +934,19 @@ def bump_build(version, token="build"):
     >>> semver.bump_build('3.4.5-rc.1+build.9')
     '3.4.5-rc.1+build.10'
     """
-    verinfo = parse(version)
-    verinfo["build"] = _increment_string(verinfo["build"] or (token or "build") + ".0")
-    return format_version(
-        verinfo["major"],
-        verinfo["minor"],
-        verinfo["patch"],
-        verinfo["prerelease"],
-        verinfo["build"],
-    )
+    return str(VersionInfo.parse(version).bump_build(token))
 
 
+@deprecated(version="2.10.0")
 def finalize_version(version):
     """
-    Remove any prerelease and build metadata from the version.
+    Remove any prerelease and build metadata from the version string.
+
+    .. versionadded:: 2.7.9
+       Added :func:`finalize_version`
+
+    .. deprecated:: 2.10.0
+       Use :func:`semver.VersionInfo.finalize_version` instead.
 
     :param version: version string
     :return: the finalized version string
@@ -680,10 +955,36 @@ def finalize_version(version):
     >>> semver.finalize_version('1.2.3-rc.5')
     '1.2.3'
     """
-    verinfo = parse(version)
-    return format_version(verinfo["major"], verinfo["minor"], verinfo["patch"])
+    verinfo = VersionInfo.parse(version)
+    return str(verinfo.finalize_version())
 
 
+@deprecated(version="2.10.0")
+def replace(version, **parts):
+    """
+    Replace one or more parts of a version and return the new string.
+
+    .. versionadded:: 2.9.0
+       Added :func:`replace`
+
+    .. deprecated:: 2.10.0
+       Use :func:`semver.VersionInfo.replace` instead.
+
+    :param str version: the version string to replace
+    :param dict parts: the parts to be updated. Valid keys are:
+      ``major``, ``minor``, ``patch``, ``prerelease``, or ``build``
+    :return: the replaced version string
+    :raises: TypeError, if ``parts`` contains invalid keys
+    :rtype: str
+
+    >>> import semver
+    >>> semver.replace("1.2.3", major=2, patch=10)
+    '2.2.10'
+    """
+    return str(VersionInfo.parse(version).replace(**parts))
+
+
+# ---- CLI
 def cmd_bump(args):
     """
     Subcommand: Bumps a version.
@@ -707,7 +1008,7 @@ def cmd_bump(args):
         # print the help and exit
         args.parser.parse_args(["bump", "-h"])
 
-    ver = parse_version_info(args.version)
+    ver = VersionInfo.parse(args.version)
     # get the respective method and call it
     func = getattr(ver, maptable[args.bump])
     return str(func())
@@ -739,6 +1040,19 @@ def cmd_compare(args):
     return str(compare(args.version1, args.version2))
 
 
+def cmd_nextver(args):
+    """
+    Subcommand: Determines the next version, taking prereleases into account.
+
+    Synopsis: nextver <VERSION> <PART>
+
+    :param args: The parsed arguments
+    :type args: :class:`argparse.Namespace`
+    """
+    version = VersionInfo.parse(args.version)
+    return str(version.next_version(args.part))
+
+
 def createparser():
     """
     Create an :class:`argparse.ArgumentParser` instance.
@@ -781,6 +1095,15 @@ def createparser():
     parser_check.set_defaults(func=cmd_check)
     parser_check.add_argument("version", help="Version to check")
 
+    # Create the nextver subcommand
+    parser_nextver = s.add_parser(
+        "nextver", help="Determines the next version, taking prereleases into account."
+    )
+    parser_nextver.set_defaults(func=cmd_nextver)
+    parser_nextver.add_argument("version", help="Version to raise")
+    parser_nextver.add_argument(
+        "part", help="One of 'major', 'minor', 'patch', or 'prerelease'"
+    )
     return parser
 
 
@@ -826,25 +1149,6 @@ def main(cliargs=None):
         return 2
 
 
-def replace(version, **parts):
-    """
-    Replace one or more parts of a version and return the new string.
-
-    :param str version: the version string to replace
-    :param dict parts: the parts to be updated. Valid keys are:
-      ``major``, ``minor``, ``patch``, ``prerelease``, or ``build``
-    :return: the replaced version string
-    :raises: TypeError, if ``parts`` contains invalid keys
-    :rtype: str
-
-    >>> import semver
-    >>> semver.replace("1.2.3", major=2, patch=10)
-    '2.2.10'
-    """
-    version = parse_version_info(version)
-    return str(version.replace(**parts))
-
-
 if __name__ == "__main__":
     import doctest
 
